@@ -1,7 +1,7 @@
 use v6;
 use fatal;
 
-unit class Getopt::Long:ver<0.4.0>;
+unit class Getopt::Long:ver<0.4.1>;
 
 class Exception is CORE::Exception {
 	has Str:D $.message is required;
@@ -10,17 +10,15 @@ class Exception is CORE::Exception {
 	}
 }
 
-role FormattableException is CORE::Exception {
+role FormattableException is Exception {
 	has Str:D $.format is required;
 	method new(Str $format) {
-		return self.bless(:$format);
+		my $message = $format.sprintf('some');
+		return self.bless(:$format, :$message);
 	}
 	method rethrow-with(Str $name) {
 		die Exception.new($!format.sprintf($name));
 	}
-	method message() {
-		$!format.sprintf('some');
-	}
 }
 
 class ValueInvalid does FormattableException {
@@ -29,7 +27,7 @@ class ValueInvalid does FormattableException {
 class ConverterInvalid does FormattableException {
 }
 
-sub convert(Str:D $value, Code:D $converter) {
+sub convert(Any:D $value, Code:D $converter) {
 	return $converter($value);
 	CATCH {
 		when X::Str::Numeric {
@@ -78,7 +76,7 @@ my class CountStore does Store {
 }
 
 my class ArrayStore does Store {
-	has Any:U $.type is required;
+	has Any:U $.type = Str;
 	method store-direct(Any:D $value) {
 		self.check-constraints($value);
 		$!values{$!key} //= $!type === Any ?? Array !! Array[$!type].new;
@@ -87,7 +85,7 @@ my class ArrayStore does Store {
 }
 
 my class HashStore does Store {
-	has Any:U $.type is required;
+	has Any:U $.type = Str;
 	method store-convert(Any:D $pair) {
 		my ($key, $value) = $pair.split('=', 2);
 		my $converted-value = convert($value, $!converter);
@@ -134,14 +132,18 @@ sub get-converter(Any:U $type) {
 		return &converter;
 	} elsif $type.HOW ~~ $coercion-how {
 		my &primary = get-converter($type.^constraint_type());
-		return %converter-for-type{$type} = sub coercion-converter(Str $input) {
+		return %converter-for-type{$type} = sub coercion-converter(Any $input) {
 			my $primary = primary($input);
 			return $primary ~~ $type.^target_type ?? $primary !! $type.^coerce($primary);
 		}
 	} elsif $type.HOW ~~ Metamodel::EnumHOW {
-		my $valid-values = $type.WHO.keys.sort({ $type.WHO{$^value} }).join(", ");
-		return %converter-for-type{$type} = sub enum-converter(Str $value) {
-			return $type.WHO{$value} // $type.^enum_from_value($value) // die ValueInvalid.new(qq{Can't convert %s argument "$value" to $type.^name(), valid values are: $valid-values});
+		sub valid-values() {
+			my @keys = $type.WHO.keys.sort({ $type.WHO{$^value} });
+			my @pairs = @keys.map: { sprintf('%s(%s)', $^key, $type.WHO{$^key}.value) };
+			return @pairs.join(', ');
+		}
+		return %converter-for-type{$type} = sub enum-converter(Any $value) {
+			return $type.WHO{$value} // $type.^enum_from_value($value) // die ValueInvalid.new(qq{Can't convert %s argument "$value" to $type.^name(), valid values are: &valid-values()});
 		}
 	} else {
 		die ConverterInvalid.new("No argument conversion known for %s argument (type {$type.^name})");
@@ -151,6 +153,9 @@ sub get-converter(Any:U $type) {
 role Argument {
 	has Junction:D $.constraints = all();
 }
+multi get-transformer(Argument $) {
+	return Nil;
+}
 
 role Argument::Valued does Argument {
 	has Any:U $.type = Str;
@@ -180,7 +185,14 @@ multi make-receivers(Argument::Scalar $arg, Str:D $key, @names, %values) {
 	return @names.map: { $^name => Receiver.new(:$store, :$arity, :default($arg.default)) }
 }
 
-class Argument::Array does Argument::Valued {
+role Argument::Composite does Argument::Valued {
+	has Code $.transformer;
+}
+multi get-transformer(Argument::Composite $arg) {
+	return $arg.transformer;
+}
+
+class Argument::Array does Argument::Composite {
 	has Range:D $.arity = 1..1;
 }
 multi make-receivers(Argument::Array $arg, Str:D $key, @names, %values) {
@@ -188,7 +200,8 @@ multi make-receivers(Argument::Array $arg, Str:D $key, @names, %values) {
 	return @names.map: { $^name => Receiver.new(:$store, :arity($arg.arity)) }
 }
 
-class Argument::Hash does Argument::Valued {
+
+class Argument::Hash does Argument::Composite {
 }
 multi make-receivers(Argument::Hash $arg, Str:D $key, @names, %values) {
 	my $store = HashStore.new(:$key, :type($arg.type), :converter($arg.converter), :constraints($arg.constraints), :%values);
@@ -196,13 +209,18 @@ multi make-receivers(Argument::Hash $arg, Str:D $key, @names, %values) {
 }
 
 class Argument::Counter does Argument {
+	has Any:U $.type = Int;
+	has Code:D $.converter = get-converter($!type);
 	has Bool:D $.argumented = False;
 }
 multi make-receivers(Argument::Counter $arg, Str:D $key, @names, %values) {
-	my $store = CountStore.new(:$key, :converter(get-converter(Int)), :constraints($arg.constraints), :%values);
+	my $store = CountStore.new(:$key, :converter($arg.converter), :constraints($arg.constraints), :%values);
 	my $arity = $arg.argumented ?? 0..1 !! 0..0;
 	return @names.map: { $^name => Receiver.new(:$store, :$arity, :1default) }
 }
+multi get-transformer(Argument::Counter $arg) {
+	return $arg.converter;
+}
 
 my rule name { [\w+]+ % '-' | '?' }
 
@@ -223,6 +241,7 @@ class Option {
 }
 
 class Ordered {
+	has Str:D $.name = 'some';
 	has Any:U $.type = Str;
 	has Code:D $.converter = get-converter($!type);
 	method type-name() {
@@ -241,7 +260,7 @@ method !options {
 	return @!options;
 }
 
-method new-from-objects(Getopt::Long:U: @options, @positionals) {
+method new-from-objects(Getopt::Long:U: @options, @positionals?) {
 	return self.bless(:@options, :@positionals);
 }
 
@@ -341,7 +360,7 @@ our sub parse-option(Str $pattern) {
 sub make-positional($type, $name) {
 	CATCH { when ConverterInvalid { .rethrow-with($name); }}
 	my $converter = get-converter($type);
-	return Ordered.new(:$type, :$converter);
+	return Ordered.new(:$name, :$type, :$converter);
 }
 
 my Str @ordinals = <first second third fourth fifth sixth seventh eighth nineth tenth some some> ... *;
@@ -405,20 +424,26 @@ multi sub trait_mod:<is>(Parameter $param where $param.positional, Ordered:D :$o
 }
 multi sub trait_mod:<is>(Parameter $param where $param.positional, Str:D :$option!) is export(:DEFAULT, :traits) {
 	CATCH { when ConverterInvalid { .rethrow-with("parameter $param.name()") }}
-	my $argument = Ordered.new(:type(type-for-format($option)));
+	my $argument = Ordered.new(:name($param.usage-name), :type(type-for-format($option)));
 	return $param does Formatted::Positional($argument);
 }
 multi sub trait_mod:<is>(Parameter $param where $param.positional, Code:D :option($converter)!) is export(:DEFAULT, :traits) {
-	my $argument = Ordered.new(:type($param.type), :$converter);
+	my $argument = Ordered.new(:name($param.usage-name), :type($param.type), :$converter);
 	return $param does Formatted::Positional($argument);
 }
 
+role Custom {
+	method get-argument-for(Parameter $param) { ... }
+}
+
 multi get-argument(Parameter $param) {
 	if $param.sigil eq '$' {
 		my $type = $param.type;
 		my $constraints = $param.constraints;
 		if $type === Bool {
 			return Argument::Boolean.new(:$constraints, :negatable(?$param.default));
+		} elsif $type ~~ Custom {
+			return $type.get-argument-for($param);
 		} else {
 			return Argument::Scalar.new(:$type, :$constraints);
 		}
@@ -438,13 +463,6 @@ sub make-option(Parameter $param) {
 	return Option.new(:names($param.named_names), :argument(get-argument($param)));
 }
 
-multi get-option-objects(&candidate) {
-	return &candidate.signature.params.grep(*.named).map(&make-option);
-}
-multi get-option-objects(&candidate where Parsed) {
-	return &candidate.getopt!options;
-}
-
 multi get-positional-object(Parameter $parameter) {
 	return make-positional($parameter.type, $parameter.name);
 }
@@ -452,29 +470,49 @@ multi get-positional-object(Parameter $param where Formatted::Positional) {
 	return $param.argument;
 }
 
-multi get-positional-objects(&candidate) {
-	return &candidate.signature.params.grep(*.positional).map(&get-positional-object);
+multi get-from-sub(&candidate) {
+	my @params = &candidate.signature.params;
+	my @options = @params.grep(*.named).map(&make-option);
+	my @positionals = @params.grep(*.positional).map(&get-positional-object);
+	return Getopt::Long.bless(:@options, :@positionals);
 }
-multi get-positional-objects(&candidate where Parsed) {
-	return &candidate.getopt!positionals;
+multi get-from-sub(&candidate where Parsed) {
+	return &candidate.getopt;
+}
+
+sub merge-named-objects(@options-for) {
+	my %unique;
+	for @options-for -> @option-set {
+		for @option-set -> $option {
+			for $option.names -> $name {
+				die "Can't merge unequal options for --$name" if %unique{$name}:exists and %unique{$name} !eqv $option;
+				%unique{$name} = $option;
+			}
+		}
+	}
+	return %unique.values;
 }
 
 sub merge-positional-object(@positionals-for, $elems) {
 	my @positionals = @positionals-for.grep(* > $elems)»[$elems];
-	die Exception.new("@ordinals[$elems].tc() arguments are of different types: { @positionals».type-name.join(', ') }") unless [eqv] @positionals;
+	die Exception.new("@ordinals[$elems].tc() arguments are of different types: { @positionals».type-name.join(', ') }") unless [eqv] @positionals».converter;
 	return @positionals[0];
 }
 
-sub merge-positional-objects(Sub $main) {
-	my @positionals-for = $main.candidates.map(&get-positional-objects);
-	my $elem-max = max(@positionals-for».elems);
+sub merge-positional-objects(@positionals-for) {
+	my $elem-max = @positionals-for».elems.max;
 	return (^$elem-max).map: { merge-positional-object(@positionals-for, $^elem) };
 }
 
+sub merge-parsers(Getopt::Long @parsers) {
+	my @options = merge-named-objects(@parsers»!options);
+	my @positionals = merge-positional-objects(@parsers»!positionals);
+	return Getopt::Long.bless(:@options, :@positionals);
+}
+
 method new-from-sub(Getopt::Long:U: Sub $main) {
-	my @options = $main.candidates.flatmap(&get-option-objects);
-	my @positionals = merge-positional-objects($main);
-	return self.bless(:@options, :@positionals);
+	my Getopt::Long @parsers = $main.candidates.map(&get-from-sub);
+	return @parsers > 1 ?? merge-parsers(@parsers) !! @parsers[0];
 }
 
 method get-options(Getopt::Long:D: @args is copy, :%hash, :$auto-abbreviate = False, :$compat-builtin = False, :named-anywhere(:$permute) = !$compat-builtin, :$bundling = !$compat-builtin, :$compat-singles = $compat-builtin, :$compat-negation = $compat-builtin, :$compat-positional = $compat-builtin, :$compat-space = $compat-builtin, :$auto-help = $compat-builtin, :$write-args) {
@@ -581,11 +619,24 @@ method get-options(Getopt::Long:D: @args is copy, :%hash, :$auto-abbreviate = Fa
 		}
 	}
 
+	sub transformer(Option $option) {
+		my $transformer = get-transformer($option.argument);
+		return $transformer ?? ($option.key => $transformer) !! ();
+	}
+	for @!options.flatmap(&transformer) -> (:$key, :$value) {
+		CATCH { when ValueInvalid { .rethrow-with("--$key") } }
+		%hash{$key} = convert(%hash{$key}, $value) if %hash{$key}:exists;
+	}
+
 	my &fallback-converter = $compat-positional ?? &val !! *.self;
-	my @converters = |@!positionals».converter, &fallback-converter, *;
-	my @positionals = (@list Z @ordinals Z @converters).map: -> $ [ $value, $name, $converter ] {
-		CATCH { when ValueInvalid { .rethrow-with($name) }}
-		convert($value, $converter);
+	my @positionals = @list.kv.map: -> $index, $value {
+		with @!positionals[$index] -> $positional {
+			CATCH { when ValueInvalid { .rethrow-with($positional.name) }}
+			convert($value, $positional.converter);
+		} else {
+			CATCH { when ValueInvalid { .rethrow-with(@ordinals[$index]) }}
+			convert($value, &fallback-converter);
+		}
 	};
 
 	@$write-args = @list if $write-args;
@@ -595,15 +646,15 @@ method get-options(Getopt::Long:D: @args is copy, :%hash, :$auto-abbreviate = Fa
 our sub get-options-from(@args, *@elements, :$overwrite, *%config) is export(:DEFAULT, :functions) {
 	my %hash := @elements && @elements[0] ~~ Hash ?? @elements.shift !! {};
 	my @options;
-	for @elements -> $element {
-		when $element ~~ Str {
-			@options.push: $element;
+	for @elements {
+		when Str {
+			@options.push: parse-option($_);
 		}
-		when $element ~~ Pair {
-			my $key = $element.key;
-			my ($name) = $element.key ~~ / ^ (\w+) /[0];
-			%hash{$name} := $element.value;
-			given $element.value {
+		when Pair {
+			my $key = .key;
+			my ($name) = .key ~~ / ^ (\w+) /[0];
+			%hash{$name} := .value;
+			given .value {
 				when Positional {
 					$key ~= '@' unless $key.ends-with('@'|'}');
 				}
@@ -611,13 +662,16 @@ our sub get-options-from(@args, *@elements, :$overwrite, *%config) is export(:DE
 					$key ~= '%' unless $key.ends-with('%');
 				}
 			}
-			@options.push: $key;
+			@options.push: parse-option($key);
+		}
+		when Option {
+			@options.push: $_;
 		}
 		default {
-			die Exception.new("Unknown element type: " ~ $element.perl);
+			die Exception.new("Unknown element type: " ~ .perl);
 		}
 	}
-	my $getopt = Getopt::Long.new-from-patterns(@options);
+	my $getopt = Getopt::Long.new-from-objects(@options);
 	return $getopt.get-options(@args, |%config, :%hash, :write-args($overwrite ?? @args !! Any));
 }
 
@@ -626,13 +680,13 @@ our sub get-options(|args) is export(:DEFAULT, :functions) {
 }
 
 our sub call-with-getopt(&func, @args, %options = %*SUB-MAIN-OPTS // {}) is export(:DEFAULT, :functions) {
-	my $capture = Getopt::Long.new-from-sub(&func).get-options(@args, |%options, :write-args(@args));
+	my $capture = Getopt::Long.new-from-sub(&func).get-options(@args, |%options);
 	return func(|$capture);
 }
 
 our sub ARGS-TO-CAPTURE(Sub $func, @args) is export(:DEFAULT, :MAIN) {
 	my %options = %*SUB-MAIN-OPTS // {};
-	return Getopt::Long.new-from-sub($func).get-options(@args, |%options, :write-args(@args));
+	return Getopt::Long.new-from-sub($func).get-options(@args, |%options);
 	CATCH { when Exception { note .message; &*EXIT(2) } };
 }
 
